Attribute VB_Name = "Parser"
'**********************************************************
'***     Standard module with procedures for parsing    ***
'***     and debugging a program.                       ***
'***     Part of the RodRego application.               ***
'**********************************************************
'***     Programmer : Nikolai Shvertner                 ***
'***     Date       : 6/20/2001                         ***
'***     Compiler   : Visual Basic 5                    ***
'***     Notes      :                                   ***
'***     Revisions  :                                   ***
'**********************************************************
Option Explicit

' Convenience constants for Error level returns
Global Const SUCCESS = 0
Global Const ERROR_LINE_NUMBER_EXPECTED = 12
Global Const ERROR_UNEXPECTED_END_OF_LINE = 13
Global Const ERROR_INVALID_OPERATION = 14
Global Const ERROR_REGISTER_EXPECTED = 15
Global Const ERROR_INVALID_LINE_ARGUMENT = 16
Global Const ERROR_TOO_MANY_ARGUMENTS = 17
Global Const ERROR_NO_PROGRAM = 18
Global Const ERROR_DUPLICATE_LINE = 19
Global Const ERROR_RETURN = -1
Global Const TOKEN_NLCR = "nlcr9754367899"
Global Const TOKEN_EOF = "eof7545425214"

' A custom defined data type for error level returns
Public Type error
    error_code As Byte  ' The code of the error
    line As Byte        ' The line where the error occurred
End Type

Function ParseString(strProg As Variant) As error
    ' A classic parser function for a simple programming language
    Dim ch As String
    Dim tokens(0 To 1000) As String
    Dim pointer, line, ptToken, i, line_prog, op, reg, l1, l2 As Integer
    
    ' Clear the memory
    For i = 0 To 255
        Program(i).OpCode = OPCODE_NONE
    Next i
    
    ' Initilize variables
    pointer = 1 ' Marker for reading the text
    ptToken = 0 ' Token index
    tokens(0) = TOKEN_EOF ' Initial token configuration
    line = 0    ' Line counter
    
  '### Phase I: Break the text into Tokens (strings)
  
  While (pointer <= Len(strProg))
  
    ' Skip all white space characters (spaces)
    ch = Mid(strProg, pointer, 1)
    While (pointer <= Len(strProg) And ch = " ")
        pointer = pointer + 1
        If (pointer <= Len(strProg)) Then ch = Mid(strProg, pointer, 1)
    Wend
    
    ' Clear the token
    tokens(ptToken) = ""
    
    ' Feed characters in the token, until encountering a white
    ' space character
    While (pointer <= Len(strProg) And ch <> " " And ch <> Chr(13))
        tokens(ptToken) = tokens(ptToken) & ch
        pointer = pointer + 1
        If (pointer <= Len(strProg)) Then ch = Mid(strProg, pointer, 1)
    Wend
    
    ' Check if something got in the Token
    If (tokens(ptToken) <> "") Then ptToken = ptToken + 1
    
    If (pointer <= Len(strProg) And ch = Chr(13)) Then
        ' A new line was encountered, so
        ' advance  the required number of characters
        ' and make a NEW-LINE Token
        pointer = pointer + 2
        tokens(ptToken) = TOKEN_NLCR
        ptToken = ptToken + 1
    End If
    
    If (pointer > Len(strProg)) Then
        ' Reached the end of the text. Make a special Token.
        tokens(ptToken) = TOKEN_EOF
        ptToken = ptToken + 1
    End If
  Wend ' Close the loop
    
  ptToken = 0 ' Reset the Token index pointer
    
  '### Phase II: Analyze the Tokens and extract the meaning
  
  ' Until we reach the final Token...
  While (tokens(ptToken) <> TOKEN_EOF)
        ' Skip new-line Tokens and Comment Tokens
        While (tokens(ptToken) = TOKEN_NLCR Or IsComment(tokens(ptToken)))
            If (IsComment(tokens(ptToken))) Then
                ' We've got a comment. Skip till the end of the line.
                While (tokens(ptToken) <> TOKEN_NLCR And tokens(ptToken) <> TOKEN_EOF)
                    ptToken = ptToken + 1
                Wend
            Else
                ptToken = ptToken + 1
                line = line + 1
            End If
        Wend
        
        If (tokens(ptToken) = TOKEN_EOF) Then
            ' The end of the Token sequence is reached.
            ' Mission complete.
            ParseString.error_code = SUCCESS
            Exit Function
        End If
        
        ' Try to convert the Token to a number and advance to
        ' the next Token.
        line_prog = String2Num(tokens(ptToken))
        ptToken = ptToken + 1
        
        If (line_prog = ERROR_RETURN) Then
            ' Problem: the first Token on a line must be a number
            ParseString.error_code = ERROR_LINE_NUMBER_EXPECTED
            ParseString.line = line
            Exit Function
        End If
        
        ' Try to obtain the code of the Operand
        op = getOpCode(tokens(ptToken))
        ptToken = ptToken + 1
        
        If (op = ERROR_RETURN) Then
            ' Problem: Couldn't recognize the operation string
            ParseString.error_code = ERROR_INVALID_OPERATION
            ParseString.line = line
            Exit Function
        End If
        
        ' Try to acquire the arguments as needed
        If (op <> OPCODE_END) Then
            reg = String2Num(tokens(ptToken))
            ptToken = ptToken + 1
            
            If (reg = ERROR_RETURN) Then
                ParseString.error_code = ERROR_REGISTER_EXPECTED
                ParseString.line = line
                Exit Function
            End If
            
            l1 = String2Num(tokens(ptToken))
            ptToken = ptToken + 1
            
            If (l1 = ERROR_RETURN) Then
                ParseString.error_code = ERROR_INVALID_LINE_ARGUMENT
                ParseString.line = line
                Exit Function
            End If
            
            If (op = OPCODE_DEB) Then
                l2 = String2Num(tokens(ptToken))
                ptToken = ptToken + 1
            
                If (l2 = ERROR_RETURN) Then
                    ParseString.error_code = ERROR_INVALID_LINE_ARGUMENT
                    ParseString.line = line
                    Exit Function
                End If
            End If
        End If
        
        If Program(line_prog).OpCode <> OPCODE_NONE Then
            ' Problem: The line already exists...
            ParseString.error_code = ERROR_DUPLICATE_LINE
            ParseString.line = line
            Exit Function
        End If
        
        ' Encode the command line (statement encoding)
        Select Case (op)
            Case OPCODE_END
                Program(line_prog).OpCode = op
            Case OPCODE_INC
                Program(line_prog).OpCode = op
                Program(line_prog).Register = reg
                Program(line_prog).Line1 = l1
            Case OPCODE_DEB
                Program(line_prog).OpCode = op
                Program(line_prog).Register = reg
                Program(line_prog).Line1 = l1
                Program(line_prog).Line2 = l2
        End Select
        
        ' Only a comment is acceptable until the end of the line
        If (IsComment(tokens(ptToken))) Then
            ' All right: skip till the end of the line
            While (tokens(ptToken) <> TOKEN_NLCR And tokens(ptToken) <> TOKEN_EOF)
                ptToken = ptToken + 1
            Wend
        Else
            If (tokens(ptToken) <> TOKEN_NLCR And tokens(ptToken) <> TOKEN_EOF) Then
                ' Problem: Non-comment Token
                ParseString.error_code = ERROR_TOO_MANY_ARGUMENTS
                ParseString.line = line
                Exit Function
            End If
        End If
  Wend
  
  ' Done
  ParseString.error_code = SUCCESS
End Function

Function ValidLetter(strChar As Variant)
    ' If the letter in the argument is in the range of
    ' the alphabet characters, return TRUE
    
    If strChar >= "A" And strChar <= "Z" Then
        ValidLetter = True
        Exit Function
    End If
    
    If strChar >= "a" And strChar <= "z" Then
        ValidLetter = True
        Exit Function
    End If
    
    ' Hmmm, that's not a letter
    ValidLetter = False
End Function

Function ValidDigit(strChar As Variant)
    ' is TRUE if a character is in the range of valid digits
    If strChar >= "0" And strChar <= "9" Then
        ValidDigit = True
        Exit Function
    End If
    
    ValidDigit = False
End Function

Function ValidNumber(strNumber As Variant)
    ' TRUE if the string is contains only digits
    Dim i As Integer
    
    For i = 1 To Len(strNumber)
        If ValidDigit(Mid(strNumber, i, 1)) = False Then
            ' The character failed the test, so reject
            ValidNumber = False
            Exit Function
        End If
    Next i
    
    ValidNumber = True
End Function

Function Character2Digit(strChar As Variant)
    ' Convert a valid digit character to its numeric value
    
    If ValidDigit(strChar) Then
        ' The character is a digit - convert it
        Character2Digit = Asc(strChar) - Asc("0")
        Exit Function
    End If
    
    ' Otherwise e are in trouble
    Character2Digit = ERROR_RETURN
End Function

Function String2Num(strNumber As Variant)
    ' Converts a string to its numeric representation
    
    Dim i As Integer
    
    If ValidNumber(strNumber) = False Then
        ' The string is not a valid number
        String2Num = ERROR_RETURN
        Exit Function
    End If
      
    ' Convert the string to the number digit by digit
    String2Num = 0
    For i = 1 To Len(strNumber)
        String2Num = String2Num * 10 + Character2Digit(Mid(strNumber, i, 1))
    Next i
End Function

Function IsComment(strNumber As Variant)
    ' A string is a comment, if its first character is a pound
    If (Len(strNumber) = 0 Or Mid(strNumber, 1, 1) <> "#") Then
        IsComment = False
    Else
        IsComment = True
    End If
End Function

Function getOpCode(strNumber As Variant)
    ' A case insensitive check, if a string is a
    ' one of the three instructions
    Select Case (UCase(strNumber))
        Case INSTR_INC
            getOpCode = OPCODE_INC
        Case INSTR_DEB
            getOpCode = OPCODE_DEB
        Case INSTR_END
            getOpCode = OPCODE_END
        Case Else
            getOpCode = ERROR_RETURN
    End Select
End Function

Function getError(ByVal error_code As Integer)
    ' Convert an error code to a string message
    Select Case error_code
        Case SUCCESS
            getError = "OK"
        Case ERROR_LINE_NUMBER_EXPECTED
            getError = "Line Number Expected"
        Case ERROR_UNEXPECTED_END_OF_LINE
            getError = "Unexpected End of Line"
        Case ERROR_INVALID_OPERATION
            getError = "Invalid Operation"
        Case ERROR_REGISTER_EXPECTED
            getError = "Valid Register Expected"
        Case ERROR_INVALID_LINE_ARGUMENT
            getError = "Invalid Argument"
        Case ERROR_TOO_MANY_ARGUMENTS
            getError = "Too Many Arguments"
        Case ERROR_NO_PROGRAM
            getError = "Empty Program"
        Case ERROR_DUPLICATE_LINE
            getError = "Duplicate Line"
        Case Else
            getError = ""
    End Select
End Function

Function VerifyProg(strProg As Variant) As error
    ' An additional level of Program syntax analysis
    
    Dim i, k As Integer
    Dim END_exists, flag1, flag2 As Boolean
    
    END_exists = False
    lenProgPointers = 0
    
    ' Record the indexes of all valid program lines
    For i = 0 To 255
        If Program(i).OpCode <> OPCODE_NONE Then
            ProgPointers(lenProgPointers) = i
            lenProgPointers = lenProgPointers + 1
        End If
    Next i
    
    If lenProgPointers = 0 Then
        ' Problem: No valid program line
        VerifyProg.error_code = ERROR_NO_PROGRAM
        VerifyProg.line = 0
        Exit Function
    End If
    
    ' Now analyze the line arguments of each statement.
    ' Check for arguments to non-existing lines.
    For i = 0 To lenProgPointers - 1
        Select Case Program(ProgPointers(i)).OpCode
            Case OPCODE_END
                END_exists = True
            Case OPCODE_INC
                flag1 = False
                For k = 0 To lenProgPointers - 1
                    If Program(ProgPointers(i)).Line1 = ProgPointers(k) Then
                        flag1 = True
                        k = lenProgPointers - 1
                    End If
                Next k
                If flag1 = False Then
                    VerifyProg.error_code = ERROR_INVALID_LINE_ARGUMENT
                    VerifyProg.line = i
                    Exit Function
                End If
                If Program(ProgPointers(i)).Register > 12 Then
                    VerifyProg.error_code = ERROR_REGISTER_EXPECTED
                    VerifyProg.line = i
                    Exit Function
                End If
            Case OPCODE_DEB
                flag1 = False
                flag2 = False
                For k = 0 To lenProgPointers - 1
                    If Program(ProgPointers(i)).Line1 = ProgPointers(k) Then
                        flag1 = True
                        If (flag2 = True) Then k = lenProgPointers - 1
                    End If
                    If Program(ProgPointers(i)).Line2 = ProgPointers(k) Then
                        flag2 = True
                        If (flag1 = True) Then k = lenProgPointers - 1
                    End If
                Next k
                If flag1 = False Or flag2 = False Then
                    VerifyProg.error_code = ERROR_INVALID_LINE_ARGUMENT
                    VerifyProg.line = i
                    Exit Function
                End If
                If Program(ProgPointers(i)).Register > 12 Then
                    VerifyProg.error_code = ERROR_REGISTER_EXPECTED
                    VerifyProg.line = i
                    Exit Function
                End If
            Case Else
                    VerifyProg.error_code = ERROR_INVALID_OPERATION
                    VerifyProg.line = i
                    Exit Function
        End Select
    Next i
    
    ' Done
    VerifyProg.error_code = SUCCESS
End Function
